from typing import Sequence, List, Dict, Any, Optional, Union

from tensorflow.keras.layers import Dropout
import pandas as pd

from sknlp.layers import (
    MLPLayer, BertTokenizationLayer, BertLayer, BertPreprocessingLayer
)
from sknlp.vocab import Vocab
from sknlp.data import BertClassificationDataset
from ..text2vec import Bert2vec
from .deep_classifier import DeepClassifier


class BertClassifier(DeepClassifier):

    def __init__(
        self,
        classes: Sequence[str],
        is_multilabel: bool = True,
        segmenter: str = None,
        embedding_size: int = 100,
        max_sequence_length: int = 512,
        num_fc_layers: int = 2,
        fc_hidden_size: int = 256,
        output_dropout: float = 0.5,
        text2vec: Optional[Bert2vec] = None,
        **kwargs
    ):
        super().__init__(classes,
                         is_multilabel=is_multilabel,
                         segmenter=segmenter,
                         algorithm="bert_classifier",
                         embedding_size=embedding_size,
                         max_sequence_length=max_sequence_length,
                         text2vec=text2vec,
                         **kwargs)
        self.mlp_layer = MLPLayer(num_fc_layers,
                                  hidden_size=fc_hidden_size,
                                  output_size=self.num_classes,
                                  name="mlp")
        self._output_dropout = output_dropout

    def create_dataset_from_df(
        self,
        df: pd.DataFrame,
        vocab: Vocab,
        segmenter: str,
        labels: Sequence[str]
    ) -> BertClassificationDataset:
        return BertClassificationDataset(
            vocab,
            list(labels),
            df=df,
            is_multilabel=self._is_multilabel,
            max_length=self._max_sequence_length,
            text_segmenter=segmenter
        )

    def build_encode_layer(self, inputs):
        _, cls = inputs
        if self._output_dropout:
            return Dropout(self._output_dropout)(cls)
        return cls

    def build_output_layer(self, inputs):
        return self.mlp_layer(inputs)

    def get_config(self):
        return {**super().get_config(), "output_dropout": self._output_dropout}

    @property
    def output_names(self) -> List[str]:
        return ["mlp"]

    @property
    def output_types(self) -> List[str]:
        return ["float"]

    @property
    def output_shapes(self) -> List[List[int]]:
        return [[-1, self.num_classes]]

    def get_custom_objects(self):
        return {
            **super().get_custom_objects(),
            "MLPLayer": MLPLayer,
            "BertTokenizationLayer": BertTokenizationLayer,
            "BertLayer": BertLayer,
            "BertPreprocessingLayer": BertPreprocessingLayer
        }



from typing import Optional, List, Dict, Any
import os
from collections import Counter

import tensorflow as tf
from tensorflow.keras.initializers import TruncatedNormal
from official.modeling import activations

from sknlp.vocab import Vocab
from sknlp.layers.bert_layer import BertLayer, BertPreprocessingLayer
from sknlp.layers.bert_tokenization import BertTokenizationLayer

from .text2vec import Text2vec


class Bert2vec(Text2vec):

    def __init__(
        self,
        vocab: Vocab,
        segmenter: Optional[str] = None,
        hidden_size: int = 768,
        num_layers: int = 12,
        num_attention_heads: int = 12,
        sequence_length: int = 512,
        max_sequence_length: int = 512,
        type_vocab_size: int = 16,
        intermediate_size: int = 3072,
        activation: activations = activations.gelu,
        dropout_rate: float = 0.1,
        attention_dropout_rate: float = 0.1,
        initializer: tf.keras.initializers.Initializer = TruncatedNormal(stddev=0.02),
        return_all_encoder_outputs: bool = False,
        only_output_cls = False,
        name: str = "bert2vec",
        **kwargs
    ) -> None:
        super().__init__(
            vocab,
            segmenter=segmenter,
            sequence_length=sequence_length,
            max_sequence_length=max_sequence_length,
            name=name,
            **kwargs
        )
        self._hidden_size = hidden_size
        self._only_output_cls = only_output_cls
        bert_preprocessing_layer = BertPreprocessingLayer(
            self.vocab.sorted_tokens[2:], sequence_length
        )
        bert_layer = BertLayer(
            len(self.vocab),
            hidden_size=hidden_size,
            num_layers=num_layers,
            num_attention_heads=num_attention_heads,
            sequence_length=sequence_length,
            max_sequence_length=max_sequence_length,
            type_vocab_size=type_vocab_size,
            intermediate_size=intermediate_size,
            activation=activation,
            dropout_rate=dropout_rate,
            attention_dropout_rate=attention_dropout_rate,
            initializer=initializer,
            return_all_encoder_outputs=return_all_encoder_outputs,
            name="bert_layer"
        )
        inputs = tf.keras.Input(shape=(1,), dtype=tf.string, name="text_input")
        if only_output_cls:
            _, outputs = bert_layer(bert_preprocessing_layer(inputs))
        else:
            outputs = bert_layer(bert_preprocessing_layer(inputs))
        self._model = tf.keras.Model(inputs=inputs, outputs=outputs, name=name)

    @property
    def input_names(self) -> List[str]:
        return ["text_input"]

    @property
    def input_types(self) -> List[str]:
        return ["string"]

    @property
    def input_shapes(self) -> List[List[int]]:
        return [[-1, 1]]

    @property
    def output_names(self) -> List[str]:
        if self._only_output_cls:
            return ["bert_layer"]
        return ["bert_layer", "bert_layer_1"]

    @property
    def output_types(self) -> List[str]:
        if self._only_output_cls:
            return ["float"]
        return ["float", "float"]

    @property
    def output_shapes(self) -> List[List[int]]:
        if self._only_output_cls:
            return [[-1, self.embedding_size]]
        sequence_length = self.sequence_length or -1
        return [[-1, sequence_length, self.embedding_size], [-1, self.embedding_size]]

    @property
    def embedding_size(self):
        return self._hidden_size

    def get_config(self) -> Dict[str, Any]:
        return {**super().get_config(),
                "hidden_size": self.embedding_size,
                "only_output_cls": self._only_output_cls}

    @classmethod
    def from_tfv1_checkpoint(
        cls, v1_checkpoint, only_output_cls=False, sequence_length=None
    ):
        bert_layer = BertLayer.from_tfv1_checkpoint(v1_checkpoint, sequence_length)
        with open(os.path.join(v1_checkpoint, "vocab.txt")) as f:
            vocab_list = f.read().split("\n")
            bert_preprocessing_layer = BertPreprocessingLayer(
                vocab_list[2:], sequence_length
            )
            vocab = Vocab(
                Counter(vocab_list), pad_token=vocab_list[0], unk_token=vocab_list[1]
            )

        module = cls(
            vocab,
            hidden_size=bert_layer.hidden_size,
            num_layers=bert_layer.num_layers,
            num_attention_heads=bert_layer.num_attention_heads,
            sequence_length=bert_layer.sequence_length,
            max_sequence_length=bert_layer.max_sequence_length,
            type_vocab_size=bert_layer.type_vocab_size,
            intermediate_size=bert_layer.intermediate_size,
            activation=bert_layer.activation,
            dropout_rate=bert_layer.dropout_rate,
            attention_dropout_rate=bert_layer.attention_dropout_rate,
            initializer=bert_layer.initializer,
            return_all_encoder_outputs=bert_layer.return_all_encoder_outputs,
            only_output_cls=only_output_cls
        )
        inputs = tf.keras.Input(shape=(1,), dtype=tf.string, name="text_input")
        if only_output_cls:
            _, outputs = bert_layer(bert_preprocessing_layer(inputs))
        else:
            outputs = bert_layer(bert_preprocessing_layer(inputs))
        module._model = tf.keras.Model(inputs=inputs, outputs=outputs, name="bert2vec")
        return module

    def get_custom_objects(self) -> Dict[str, Any]:
        return {
            **super().get_custom_objects(),
            "BertTokenizationLayer": BertTokenizationLayer,
            "BertLayer": BertLayer,
            "BertPreprocessingLayer": BertPreprocessingLayer,
        }



from .mlp import MLPLayer
from .lstmp import LSTMPCell, LSTMP
from .multi_lstmp import MultiLSTMP
from .bert_tokenization import BertTokenizationLayer
from .bert_layer import BertLayer


__all__ = [
    "MLPLayer", "LSTMPCell", "LSTMP", "MultiLSTMP",
    "BertTokenizationLayer", "BertLayer", "BertPreprocessingLayer"
]



from typing import List, Optional, Dict, Any

from sknlp.layers import MultiLSTMP, MLPLayer, LSTMP
from .deep_classifier import DeepClassifier


class TextRNNClassifier(DeepClassifier):

    def __init__(self,
                 classes: List[str],
                 is_multilabel: bool = True,
                 max_sequence_length: int = 100,
                 sequence_length: int = None,
                 segmenter: str = "jieba",
                 embedding_size: int = 100,
                 num_rnn_layers: int = 1,
                 rnn_hidden_size: int = 512,
                 rnn_projection_size: int = 128,
                 rnn_recurrent_clip: float = 3.0,
                 rnn_projection_clip: float = 3.0,
                 rnn_input_dropout: float = 0.5,
                 rnn_recurrent_dropout: float = 0.5,
                 rnn_output_dropout: float = 0.5,
                 num_fc_layers: int = 2,
                 fc_hidden_size: int = 128,
                 rnn_kernel_initializer="glorot_uniform",
                 rnn_recurrent_initializer="orthogonal",
                 rnn_projection_initializer="glorot_uniform",
                 rnn_bias_initializer="zeros",
                 rnn_kernel_regularizer=None,
                 rnn_recurrent_regularizer=None,
                 rnn_projection_regularizer=None,
                 rnn_bias_regularizer=None,
                 rnn_kernel_constraint=None,
                 rnn_recurrent_constraint=None,
                 rnn_projection_constraint=None,
                 rnn_bias_constraint=None,
                 text2vec=None,
                 **kwargs):
        super().__init__(classes,
                         is_multilabel=is_multilabel,
                         max_sequence_length=max_sequence_length,
                         sequence_length=sequence_length,
                         segmenter=segmenter,
                         embedding_size=embedding_size,
                         algorithm="text_rnn",
                         **kwargs)
        self.rnn_layer = MultiLSTMP(
            num_rnn_layers,
            rnn_hidden_size,
            projection_size=rnn_projection_size,
            recurrent_clip=rnn_recurrent_clip,
            projection_clip=rnn_projection_clip,
            input_dropout=rnn_input_dropout,
            recurrent_dropout=rnn_recurrent_dropout,
            output_dropout=rnn_output_dropout,
            last_connection="last",
            kernel_initializer=rnn_kernel_initializer,
            recurrent_initializer=rnn_recurrent_initializer,
            projection_initializer=rnn_projection_initializer,
            bias_initializer=rnn_bias_initializer,
            kernel_regularizer=rnn_kernel_regularizer,
            recurrent_regularizer=rnn_recurrent_regularizer,
            projection_regularizer=rnn_projection_regularizer,
            bias_regularizer=rnn_bias_regularizer,
            kernel_constraint=rnn_kernel_constraint,
            recurrent_constraint=rnn_recurrent_constraint,
            projection_constraint=rnn_projection_constraint,
            bias_constraint=rnn_bias_constraint,
            name="rnn"
        )
        self.mlp_layer = MLPLayer(
            num_fc_layers, hidden_size=fc_hidden_size,
            output_size=self.num_classes, name="mlp"
        )

    def build_encode_layer(self, inputs):
        return self.rnn_layer(inputs)

    def build_output_layer(self, inputs):
        return self.mlp_layer(inputs)

    @property
    def output_names(self) -> List[str]:
        return ["mlp"]

    @property
    def output_types(self) -> List[str]:
        return ["float"]

    @property
    def output_shapes(self) -> List[List[int]]:
        return [[-1, self.num_classes]]

    def get_custom_objects(self):
        return {
            **super().get_custom_objects(),
            "MultiLSTMP": MultiLSTMP,
            "MLPLayer": MLPLayer,
            "LSTMP": LSTMP
        }



from typing import Sequence, List, Optional

import numpy as np
import pandas as pd

import tensorflow as tf

from sknlp.vocab import Vocab

from .nlp_dataset import NLPDataset


class BertClassificationDataset(NLPDataset):

    def __init__(
        self,
        vocab: Vocab,
        labels: Sequence[str],
        df: Optional[pd.DataFrame] = None,
        csv_file: Optional[str] = None,
        in_memory: bool = True,
        is_multilabel: bool = True,
        max_length: Optional[int] = 80,
        text_segmenter: str = None
    ):
        self.vocab = vocab
        self.labels = labels
        self.is_multilabel = is_multilabel
        self.max_length = max_length or 99999
        self.label2idx = dict(zip(labels, range(len(labels))))
        super().__init__(df=df,
                         csv_file=csv_file,
                         in_memory=in_memory,
                         text_segmenter=text_segmenter,
                         text_dtype=tf.string,
                         label_dtype=tf.float32,
                         text_padding_shape=(None,),
                         label_padding_shape=(None,),
                         text_padding_value=vocab.pad,
                         label_padding_value=0.0)

    def _text_transform(self, text: tf.Tensor) -> List[tf.Tensor]:
        return [text]

    def _label_binarizer(self, labels: List[str]) -> np.ndarray:
        label2idx = self.label2idx
        res = np.zeros(len(label2idx), dtype=np.float32)
        res[[label2idx[label] for label in labels if label in label2idx]] = 1
        return res

    def _label_transform(self, label: tf.Tensor) -> np.ndarray:
        label = super()._label_transform(label)
        if self.is_multilabel:
            labels = label.split('|')
        else:
            labels = [label]
        return self._label_binarizer(labels)

    def batchify(
        self,
        batch_size: int,
        shuffle: bool = True,
        shuffle_buffer_size: int = 100000
    ) -> tf.data.Dataset:
        dataset = self._dataset
        if shuffle:
            if shuffle_buffer_size <= 0:
                shuffle_buffer_size = 100000
            dataset = dataset.shuffle(shuffle_buffer_size)

        return (
            dataset
            .batch(batch_size)
            .prefetch(buffer_size=tf.data.experimental.AUTOTUNE)
        )



from typing import List, Dict, Any
import os
import tempfile

import tensorflow as tf
from tensorflow.keras.initializers import TruncatedNormal
from tensorflow.keras.layers import InputSpec
from tensorflow.keras.layers.experimental.preprocessing import TextVectorization

from official.modeling import activations
from official.nlp.modeling import layers
from official.nlp.bert import configs

from .bert_tokenization import BertTokenizationLayer
from .utils import convert_bert_checkpoint, create_bert_model


class BertPreprocessingLayer(tf.keras.layers.Layer):

    def __init__(self, vocab: List[str], sequence_length: int = None, **kwargs) -> None:
        self.vocab = vocab
        self.sequence_length = sequence_length
        self.text2vec = TextVectorization(
            max_tokens=len(vocab) + 3,
            output_sequence_length=sequence_length,
            standardize=None,
            split=BertTokenizationLayer()
        )
        self.text2vec.set_vocabulary(vocab)
        super().__init__(**kwargs)

    def call(self, inputs: tf.Tensor) -> tf.Tensor:
        return self.text2vec(inputs)

    def get_config(self) -> Dict[str, Any]:
        return {
            **super().get_config(),
            "vocab": self.vocab,
            "sequence_length": self.sequence_length
        }


class BertLayer(tf.keras.layers.Layer):
    """
    copy from https://github.com/tensorflow/models/blob/v2.2.0/official/nlp/modeling/networks/transformer_encoder.py.
    """

    def __init__(
        self,
        vocab_size: int,
        hidden_size: int = 768,
        num_layers: int = 12,
        num_attention_heads: int = 12,
        sequence_length: int = None,
        max_sequence_length: int = 512,
        type_vocab_size: int = 16,
        intermediate_size: int = 3072,
        activation: activations = activations.gelu,
        dropout_rate: float = 0.1,
        attention_dropout_rate: float = 0.1,
        initializer: tf.keras.initializers.Initializer = TruncatedNormal(stddev=0.02),
        return_all_encoder_outputs: bool = False,
        name: str = "bert_layer",
        **kwargs
    ):
        self.vocab_size = vocab_size
        self.hidden_size = hidden_size
        self.num_layers = num_layers
        self.num_attention_heads = num_attention_heads
        self.sequence_length = sequence_length
        self.max_sequence_length = max_sequence_length
        self.type_vocab_size = type_vocab_size
        self.intermediate_size = intermediate_size
        self.activation = activation
        self.dropout_rate = dropout_rate
        self.attention_dropout_rate = attention_dropout_rate
        self.initializer = initializer
        self.return_all_encoder_outputs = return_all_encoder_outputs
        self.input_spec = InputSpec(ndim=2, shape=(None, self.sequence_length))

        self.embedding_layer = layers.OnDeviceEmbedding(
            vocab_size=vocab_size,
            embedding_width=hidden_size,
            initializer=initializer,
            name="word_embeddings"
        )
        # Always uses dynamic slicing for simplicity.
        self.position_embedding_layer = layers.PositionEmbedding(
            initializer=initializer,
            use_dynamic_slicing=True,
            max_sequence_length=max_sequence_length,
            name="position_embeddings"
        )
        self.type_embedding_layer = layers.OnDeviceEmbedding(
            vocab_size=type_vocab_size,
            embedding_width=hidden_size,
            initializer=initializer,
            use_one_hot=True,
            name="type_embeddings"
        )
        self.normalize_layer = tf.keras.layers.LayerNormalization(
            name="embeddings/layer_norm",
            axis=-1,
            epsilon=1e-12,
            dtype=tf.float32
        )
        self.transformer_layers = []
        for i in range(num_layers):
            layer = layers.Transformer(
                num_attention_heads=num_attention_heads,
                intermediate_size=intermediate_size,
                intermediate_activation=activation,
                dropout_rate=dropout_rate,
                attention_dropout_rate=attention_dropout_rate,
                kernel_initializer=initializer,
                name="transformer/layer_%d" % i
            )
            self.transformer_layers.append(layer)

        self.cls_output_layer = tf.keras.layers.Dense(
            units=hidden_size,
            activation="tanh",
            kernel_initializer=initializer,
            name="pooler_transform"
        )

        super().__init__(name=name, **kwargs)

    def call(self, inputs):
        input_masks = inputs != 0
        input_types = tf.zeros_like(inputs, name="input_type_ids")

        word_embeddings = self.embedding_layer(inputs)
        position_embeddings = self.position_embedding_layer(word_embeddings)
        type_embeddings = self.type_embedding_layer(input_types)
        embeddings = word_embeddings + position_embeddings + type_embeddings
        embeddings = self.normalize_layer(embeddings)
        embeddings = tf.keras.layers.Dropout(rate=self.dropout_rate)(embeddings)

        data = embeddings
        attention_mask = layers.SelfAttentionMask()([data, input_masks])
        encoder_outputs = []
        for layer in self.transformer_layers:
            data = layer([data, attention_mask])
            encoder_outputs.append(data)

        first_token_tensor = tf.keras.layers.Lambda(
            lambda x: tf.squeeze(x[:, 0:1, :], axis=1)
        )(encoder_outputs[-1])
        cls_output = self.cls_output_layer(first_token_tensor)
        if self.return_all_encoder_outputs:
            return [encoder_outputs, cls_output]
        else:
            return [encoder_outputs[-1], cls_output]

    def get_config(self):
        return {
            **super().get_config(),
            "vocab_size": self.vocab_size,
            "hidden_size": self.hidden_size,
            "num_layers": self.num_layers,
            "num_attention_heads": self.num_attention_heads,
            "sequence_length": self.sequence_length,
            "max_sequence_length": self.max_sequence_length,
            "type_vocab_size": self.type_vocab_size,
            "intermediate_size": self.intermediate_size,
            "activation": self.activation,
            "dropout_rate": self.dropout_rate,
            "attention_dropout_rate": self.attention_dropout_rate,
            "initializer": self.initializer,
            "return_all_encoder_outputs": self.return_all_encoder_outputs
        }

    @classmethod
    def from_tfv1_checkpoint(cls, v1_checkpoint, sequence_length=None):
        bert_config = configs.BertConfig.from_json_file(
            os.path.join(v1_checkpoint, "bert_config.json")
        )
        layer = cls(
            vocab_size=bert_config.vocab_size,
            hidden_size=bert_config.hidden_size,
            num_layers=bert_config.num_hidden_layers,
            num_attention_heads=bert_config.num_attention_heads,
            intermediate_size=bert_config.intermediate_size,
            activation=activations.gelu,
            dropout_rate=bert_config.hidden_dropout_prob,
            attention_dropout_rate=bert_config.attention_probs_dropout_prob,
            sequence_length=sequence_length,
            max_sequence_length=bert_config.max_position_embeddings,
            type_vocab_size=bert_config.type_vocab_size,
            initializer=tf.keras.initializers.TruncatedNormal(
                stddev=bert_config.initializer_range
            )
        )
        bert_network = create_bert_model(bert_config)
        with tempfile.TemporaryDirectory() as temp_dir:
            temporary_checkpoint = os.path.join(temp_dir, "ckpt")
            convert_bert_checkpoint(
                checkpoint_from_path=v1_checkpoint,
                checkpoint_to_path=temporary_checkpoint,
                num_heads=bert_config.num_attention_heads
            )
            bert_network.load_weights(
                temporary_checkpoint
            ).assert_existing_objects_matched()

            with open(os.path.join(v1_checkpoint, "vocab.txt")) as f:
                vocab = f.read().split("\n")
                bert_preprocessing_layer = BertPreprocessingLayer(
                    vocab[2:], sequence_length
                )

            layer(bert_preprocessing_layer(tf.constant([["是的"], ["不是的"]])))
            layer.embedding_layer.set_weights(
                bert_network.get_layer("word_embeddings").get_weights()
            )
            layer.position_embedding_layer.set_weights(
                bert_network.get_layer("position_embedding").get_weights()
            )
            layer.type_embedding_layer.set_weights(
                bert_network.get_layer("type_embeddings").get_weights()
            )
            layer.normalize_layer.set_weights(
                bert_network.get_layer("embeddings/layer_norm").get_weights()
            )
            for i in range(bert_config.num_hidden_layers):
                layer.transformer_layers[i].set_weights(
                    bert_network.get_layer("transformer/layer_%d" % i).get_weights()
                )
            layer.cls_output_layer.set_weights(
                bert_network.get_layer("pooler_transform").get_weights()
            )
            return layer



import logging

import numpy as np
import tensorflow as tf

from official.modeling import activations
from official.nlp.modeling import networks
import tensorflow.compat.v1 as tfv1


logger = logging.getLogger(__name__)
handler = logging.StreamHandler()
handler.setLevel(logging.INFO)
logger.addHandler(handler)
logger.setLevel(logging.INFO)


BERT_NAME_REPLACEMENTS = (
    ("bert/", ""),
    ("encoder", "transformer"),
    ("embeddings/word_embeddings", "word_embeddings/embeddings"),
    ("embeddings/token_type_embeddings", "type_embeddings/embeddings"),
    ("embeddings/position_embeddings", "position_embedding/embeddings"),
    ("embeddings/LayerNorm", "embeddings/layer_norm"),
    ("attention/self", "self_attention"),
    ("attention/output/dense", "self_attention_output"),
    ("attention/output/LayerNorm", "self_attention_layer_norm"),
    ("intermediate/dense", "intermediate"),
    ("output/dense", "output"),
    ("output/LayerNorm", "output_layer_norm"),
    ("pooler/dense", "pooler_transform"),
    ("cls/predictions/output_bias", "cls/predictions/output_bias/bias"),
    ("cls/seq_relationship/output_bias", "predictions/transform/logits/bias"),
    ("cls/seq_relationship/output_weights", "predictions/transform/logits/kernel"),
)
BERT_PERMUTATIONS = (("cls/seq_relationship/output_weights", (1, 0)),)


def _bert_name_replacement(var_name, name_replacements):
  """Gets the variable name replacement."""
  for src_pattern, tgt_pattern in name_replacements:
    if src_pattern in var_name:
      old_var_name = var_name
      var_name = var_name.replace(src_pattern, tgt_pattern)
      logger.info("Converted: %s --> %s", old_var_name, var_name)
  return var_name


def _has_exclude_patterns(name, exclude_patterns):
  """Checks if a string contains substrings that match patterns to exclude."""
  for p in exclude_patterns:
    if p in name:
      return True
  return False


def _get_permutation(name, permutations):
  """Checks whether a variable requires transposition by pattern matching."""
  for src_pattern, permutation in permutations:
    if src_pattern in name:
      logger.info("Permuted: %s --> %s", name, permutation)
      return permutation

  return None


def _get_new_shape(name, shape, num_heads):
  """Checks whether a variable requires reshape by pattern matching."""
  if "self_attention_output/kernel" in name:
    return (num_heads, shape[0] // num_heads, shape[1])
  if "self_attention_output/bias" in name:
    return shape

  patterns = [
      "self_attention/query", "self_attention/value", "self_attention/key"
  ]
  for pattern in patterns:
    if pattern in name:
      if "kernel" in name:
        return tuple([shape[0], num_heads, shape[1] // num_heads])
      if "bias" in name:
        return tuple([num_heads, shape[0] // num_heads])
  return None


def convert_bert_checkpoint(checkpoint_from_path, checkpoint_to_path, num_heads):
  """Migrates the names of variables within a checkpoint.
  Args:
    checkpoint_from_path: Path to source checkpoint to be read in.
    checkpoint_to_path: Path to checkpoint to be written out.
    num_heads: The number of heads of the model.
    name_replacements: A list of tuples of the form (match_str, replace_str)
      describing variable names to adjust.
    permutations: A list of tuples of the form (match_str, permutation)
      describing permutations to apply to given variables. Note that match_str
      should match the original variable name, not the replaced one.
    exclude_patterns: A list of string patterns to exclude variables from
      checkpoint conversion.
  Returns:
    A dictionary that maps the new variable names to the Variable objects.
    A dictionary that maps the old variable names to the new variable names.
  """
  name_replacements = BERT_NAME_REPLACEMENTS
  permutations = BERT_PERMUTATIONS
  exclude_patterns = ["adam", "Adam"]
  with tfv1.Graph().as_default():
    logger.info("Reading checkpoint_from_path %s", checkpoint_from_path)
    reader = tfv1.train.load_checkpoint(checkpoint_from_path)
    name_shape_map = reader.get_variable_to_shape_map()
    new_variable_map = {}
    conversion_map = {}
    for var_name in name_shape_map:
      if exclude_patterns and _has_exclude_patterns(var_name, exclude_patterns):
        continue
      # Get the original tensor data.
      tensor = reader.get_tensor(var_name)

      # Look up the new variable name, if any.
      new_var_name = _bert_name_replacement(var_name, name_replacements)

      # See if we need to reshape the underlying tensor.
      new_shape = None
      if num_heads > 0:
        new_shape = _get_new_shape(new_var_name, tensor.shape, num_heads)
      if new_shape:
        logger.info("Veriable %s has a shape change from %s to %s",
                    var_name, tensor.shape, new_shape)
        tensor = np.reshape(tensor, new_shape)

      # See if we need to permute the underlying tensor.
      permutation = _get_permutation(var_name, permutations)
      if permutation:
        tensor = np.transpose(tensor, permutation)

      # Create a new variable with the possibly-reshaped or transposed tensor.
      var = tfv1.Variable(tensor, name=var_name)

      # Save the variable into the new variable map.
      new_variable_map[new_var_name] = var

      # Keep a list of converter variables for sanity checking.
      if new_var_name != var_name:
        conversion_map[var_name] = new_var_name

    saver = tfv1.train.Saver(new_variable_map)

    with tfv1.Session() as sess:
      sess.run(tfv1.global_variables_initializer())
      logger.info("Writing checkpoint_to_path %s", checkpoint_to_path)
      saver.save(sess, checkpoint_to_path, write_meta_graph=True)

  logger.info("Summary:")
  logger.info("  Converted %d variable name(s).", len(new_variable_map))
  logger.info("  Converted: %s", str(conversion_map))



def create_bert_model(cfg):
    """Creates a BERT keras core model from BERT configuration.
    Args:
    cfg: A `BertConfig` to create the core model.
    Returns:
    A TransformerEncoder netowork.
    """
    bert_encoder = networks.TransformerEncoder(
      vocab_size=cfg.vocab_size,
      hidden_size=cfg.hidden_size,
      num_layers=cfg.num_hidden_layers,
      num_attention_heads=cfg.num_attention_heads,
      intermediate_size=cfg.intermediate_size,
      activation=activations.gelu,
      dropout_rate=cfg.hidden_dropout_prob,
      attention_dropout_rate=cfg.attention_probs_dropout_prob,
      sequence_length=80,
      max_sequence_length=cfg.max_position_embeddings,
      type_vocab_size=cfg.type_vocab_size,
      initializer=tf.keras.initializers.TruncatedNormal(stddev=cfg.initializer_range)
    )
    return bert_encoder
